/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <doca_log.h>

#include "doca_utils.h"

DOCA_LOG_REGISTER(DOCA_UTILS);

static doca_error_t
parse_bdf_from_str(const char *bdf_str, uint32_t *bus, uint32_t *device, uint32_t *function)
{
	char *endptr = NULL;

	errno = 0;
	*bus = (uint32_t)strtoul(bdf_str, &endptr, 16);
	if (errno != 0 || endptr != (bdf_str + 2) || *bus >= PCI_BUS_MAX_VALUE) {
		DOCA_LOG_ERR("Unable to parse pci bdf from string: Failed to parse the bus value");
		return DOCA_ERROR_INVALID_VALUE;
	}

	*device = (uint32_t)strtoul(bdf_str + 3, &endptr, 16);
	if (errno != 0 || endptr != (bdf_str + 5) || *device >= PCI_DEVICE_MAX_VALUE) {
		DOCA_LOG_ERR("Unable to parse pci bdf from string: Failed to parse the device value");
		return DOCA_ERROR_INVALID_VALUE;
	}

	*function = (uint32_t)strtoul(bdf_str + 6, &endptr, 16);
	if (errno != 0 || endptr != (bdf_str + 7) || *function >= PCI_FUNCTION_MAX_VALUE) {
		DOCA_LOG_ERR("Unable to parse pci bdf from string: Failed to parse the function value");
		return DOCA_ERROR_INVALID_VALUE;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_pci_bdf_from_string(const char *pci_bdf_str, struct doca_pci_bdf *pci_bdf_obj)
{
	if (pci_bdf_str == NULL || pci_bdf_obj == NULL) {
		DOCA_LOG_ERR("Unable to parse pci bdf from string: Received invalid input: pci_bdf_str or pci_bdf_obj are null");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (strnlen(pci_bdf_str, PCI_BDF_SIZE) != PCI_BDF_SIZE - 1 || pci_bdf_str[2] != ':' || pci_bdf_str[5] != '.') {
		DOCA_LOG_ERR("Unable to parse pci bdf from string: Input doesn't meet the pci address convention: XX:XX.X");
		return DOCA_ERROR_INVALID_VALUE;
	}

	uint32_t bus, device, function; /* Needed to ensure that the parsed values meet the size limits */

	doca_error_t status = parse_bdf_from_str(pci_bdf_str, &bus, &device, &function);

	if (status != DOCA_SUCCESS)
		return status; /* Error message is already printed inside parse_bdf_from_str() */

	pci_bdf_obj->bus = bus;
	pci_bdf_obj->device = device;
	pci_bdf_obj->function = function;

	return DOCA_SUCCESS;
}

doca_error_t
doca_pci_addr_from_string(const char *pci_addr_str, struct doca_pci_addr *pci_addr_obj)
{
	if (pci_addr_str == NULL || pci_addr_obj == NULL) {
		DOCA_LOG_ERR("Unable to parse pci address from string: Received invalid input: pci_addr_obj or pci_addr_str are null");
		return DOCA_ERROR_INVALID_VALUE;
	}

	size_t len = strnlen(pci_addr_str, PCI_ADDR_SIZE);

	if ((len != PCI_BDF_SIZE - 1 && len != PCI_ADDR_SIZE - 1) || pci_addr_str[len - 5] != ':' || pci_addr_str[len - 2] != '.') {
		DOCA_LOG_ERR("Unable to parse pci address from string: Input doesn't meet the pci address convention: XX:XX.X or XXXX:XX:XX:X.");
		return DOCA_ERROR_INVALID_VALUE;
	}

	const char *addr = pci_addr_str;
	char *endptr;
	/*
	 * Needed to ensure that the parsed values meet the size limits.
	 * Zeroing the domain in case the address does not include a Domain component.
	 */
	uint32_t domain = 0, bus, device, function;

	/* Extended case */
	if (len == PCI_ADDR_SIZE - 1) {
		if (pci_addr_str[4] != ':') {
			DOCA_LOG_ERR("Unable to parse pci address from string: Input doesn't meet the pci address convention: XX:XX.X or XXXX:XX:XX:X.");
			return DOCA_ERROR_INVALID_VALUE;
		}

		errno = 0;
		domain = (uint32_t)strtoul(pci_addr_str, &endptr, 16);
		if (errno != 0 || endptr != (pci_addr_str + 4) || domain >= PCI_DOMAIN_MAX_VALUE) {
			DOCA_LOG_ERR("Unable to parse pci address from string: Failed to parse the bus value");
			return DOCA_ERROR_INVALID_VALUE;
		}

		/* Adjust addr to the non-extended case (only bdf) for the rest of the method */
		addr = pci_addr_str + 5;
	}

	doca_error_t status = parse_bdf_from_str(addr, &bus, &device, &function);

	if (status != DOCA_SUCCESS)
		return status; /* Error message is already printed inside parse_bdf_from_str() */

	pci_addr_obj->domain = domain;
	pci_addr_obj->bus = bus;
	pci_addr_obj->device = device;
	pci_addr_obj->function = function;

	return DOCA_SUCCESS;
}

doca_error_t
doca_pci_addr_to_string(const struct doca_pci_addr *pci_addr_obj, char *pci_addr_str, size_t buf_size)
{
	if (pci_addr_obj == NULL || pci_addr_str == NULL || buf_size < PCI_ADDR_SIZE) {
		DOCA_LOG_ERR("Unable to generate pci address string from pci address struct: Either pci_addr_obj or pci_addr_str are null, or buf_size=%zu is too small",
			buf_size);
		return DOCA_ERROR_INVALID_VALUE;
	}

	int status = snprintf(pci_addr_str, PCI_ADDR_SIZE, "%04x:%02x:%02x.%1x", pci_addr_obj->domain,
			      pci_addr_obj->bus, pci_addr_obj->device, pci_addr_obj->function);

	if (status < 0) {
		DOCA_LOG_ERR("Unable to generate pci address string from pci address struct: Failed to write to buffer");
		return DOCA_ERROR_INVALID_VALUE;
	}

	return DOCA_SUCCESS;
}
