/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DOCA_UTILS_H_
#define DOCA_UTILS_H_

#include <doca_error.h>
#include <doca_types.h>

#ifdef __cplusplus
extern "C" {
#endif

#define PCI_BDF_SIZE 8 /* Null terminator included */
#define PCI_ADDR_SIZE 13 /* Null terminator included */

/**
 * @brief Parse a PCI BDF address string into a doca_pci_bdf struct
 *
 * @note The string format must be "XX:XX.X"
 *
 * @param [in] pci_bdf_str
 * The PCI BDF address string
 * @param [out] pci_bdf_obj
 * The doca_pci_bdf struct with the PCI BDF parameters - Bus, Device, Function
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
doca_error_t doca_pci_bdf_from_string(const char *pci_bdf_str, struct doca_pci_bdf *pci_bdf_obj);

/**
 * @brief Parse a PCI address string into a doca_pci_addr struct
 *
 * @note The two valid string formats are BDF - "XX:XX.X" and DBDF (Domain extended) - "XXXX:XX:XX.X"
 *
 * @param [in] pci_addr_str
 * The PCI address string
 * @param [out] pci_addr_obj
 * The doca_pci_addr struct with the PCI address parameters
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
doca_error_t doca_pci_addr_from_string(const char *pci_addr_str, struct doca_pci_addr *pci_addr_obj);

/**
 * @brief Generate a PCI address string from a doca_pci_addr struct
 *
 * @param [in] pci_addr_obj
 * The doca_pci_addr struct
 * @param [out] pci_addr_str
 * The generated PCI address string (see size requirement below)
 * @param [in] buf_size
 * The size of the buffer pci_addr_str, must be at least PCI_ADDR_SIZE
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
doca_error_t doca_pci_addr_to_string(const struct doca_pci_addr *pci_addr_obj, char *pci_addr_str, size_t buf_size);

#ifdef __cplusplus
}
#endif

#endif /* DOCA_UTILS_H_ */
