/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>

#include <doca_log.h>
#include <doca_error.h>

#include "flow_skeleton.h"

DOCA_LOG_REGISTER(FLOW_SKELETON);

#define DEFAULT_QUEUE_DEPTH 128

struct skeleton_ctx {
	uint32_t queue_depth;				/* DOCA Flow queue depth */
	struct flow_skeleton_cfg *skeleton_cfg;		/* pointer to skeleton config */
	uint32_t **queue_state;				/* array to monitor the queues state */
	uint32_t *queue_counter;			/* array to count how many entries processed */
	struct flow_skeleton_entry *entries;		/* array of flow skeleton entries */
	void **program_ctx;				/* application context from main loop */
};

static struct skeleton_ctx skeleton_ctx;
static volatile bool force_quit;

/*
 * Entry processing callback
 *
 * @entry [in]: entry pointer
 * @pipe_queue [in]: queue identifier
 * @status [in]: DOCA Flow entry status
 * @op [in]: DOCA Flow entry operation
 * @user_ctx [out]: user context
 */
static void
process_callback(struct doca_flow_pipe_entry *entry, uint16_t pipe_queue,
		 enum doca_flow_entry_status status, enum doca_flow_entry_op op, void *user_ctx)
{
	doca_error_t result;

	if (op == DOCA_FLOW_ENTRY_OP_ADD) {
		/* call application callback */
		skeleton_ctx.skeleton_cfg->add_cb(entry, pipe_queue, status, user_ctx, skeleton_ctx.program_ctx);
		if (status != DOCA_FLOW_ENTRY_STATUS_SUCCESS) {
			DOCA_LOG_ERR("Failed to add entry");
			/* if status is not success - the skeleton will remove the entry */
			result = doca_flow_pipe_rm_entry(pipe_queue, DOCA_FLOW_WAIT_FOR_BATCH, user_ctx, entry);
			if (result != DOCA_SUCCESS)
				DOCA_LOG_ERR("Failed to remove entry: %s", doca_get_error_string(result));
		} else
			skeleton_ctx.queue_counter[pipe_queue]++;
	} else if (op == DOCA_FLOW_ENTRY_OP_DEL) {
		/* call application callback */
		skeleton_ctx.skeleton_cfg->remove_cb(entry, pipe_queue, status, user_ctx, skeleton_ctx.program_ctx[pipe_queue]);
		if (status != DOCA_FLOW_ENTRY_STATUS_SUCCESS) {
			/* if status is not success - notify the application about a failure */
			DOCA_LOG_ERR("Failed to remove entry");
			skeleton_ctx.skeleton_cfg->failure_cb();
		} else
			skeleton_ctx.queue_counter[pipe_queue]++;
	}
}

doca_error_t
flow_skeleton_init(struct doca_flow_cfg *flow_cfg, struct flow_skeleton_cfg *skeleton_cfg)
{
	uint16_t nb_queues;
	int i;
	doca_error_t result;

	skeleton_ctx.skeleton_cfg = skeleton_cfg;
	skeleton_ctx.queue_depth = flow_cfg->queue_depth == 0 ? DEFAULT_QUEUE_DEPTH : flow_cfg->queue_depth;
	nb_queues = flow_cfg->queues;

	if (skeleton_cfg->nb_entries >= skeleton_ctx.queue_depth / 2) {
		DOCA_LOG_ERR("Queue depth should be at least twice larger than nb_entries");
		return DOCA_ERROR_INVALID_VALUE;
	}

	skeleton_ctx.entries = (struct flow_skeleton_entry *)calloc(skeleton_cfg->nb_entries, sizeof(struct flow_skeleton_entry));
	if (skeleton_ctx.entries == NULL) {
		DOCA_LOG_ERR("Failed to allocate memory");
		return DOCA_ERROR_NO_MEMORY;
	}
	skeleton_ctx.program_ctx = (void *)calloc(nb_queues, sizeof(void *));
	if (skeleton_ctx.program_ctx == NULL) {
		DOCA_LOG_ERR("Failed to allocate memory");
		result = DOCA_ERROR_NO_MEMORY;
		goto free_entries;
	}
	skeleton_ctx.queue_counter = (uint32_t *)calloc(nb_queues, sizeof(uint32_t));
	if (skeleton_ctx.queue_counter == NULL) {
		DOCA_LOG_ERR("Failed to allocate memory");
		result = DOCA_ERROR_NO_MEMORY;
		goto free_program_ctx;
	}
	skeleton_ctx.queue_state = (uint32_t **)calloc(skeleton_cfg->nb_ports, sizeof(uint32_t *));
	if (skeleton_ctx.program_ctx == NULL) {
		DOCA_LOG_ERR("Failed to allocate memory");
		result = DOCA_ERROR_NO_MEMORY;
		goto free_queue_counter;
	}
	for (i = 0; i < skeleton_cfg->nb_ports; i++) {
		skeleton_ctx.queue_state[i] = (uint32_t *)calloc(nb_queues, sizeof(uint32_t));
		if (skeleton_ctx.program_ctx == NULL) {
			int j;

			DOCA_LOG_ERR("Failed to allocate memory");
			for (j = 0; j < i; j++)
				free(skeleton_ctx.queue_state[j]);
			result = DOCA_ERROR_NO_MEMORY;
			goto free_queue_state;
		}
	}

	flow_cfg->cb = process_callback;
	result = doca_flow_init(flow_cfg);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init DOCA Flow: %s", doca_get_error_string(result));
		result = DOCA_ERROR_NO_MEMORY;
		goto free_queue_state_raw;
	}
	force_quit = false;
	return DOCA_SUCCESS;

free_queue_state_raw:
	for (i = 0; i < skeleton_cfg->nb_ports; i++)
		free(skeleton_ctx.queue_state[i]);
free_queue_state:
	free(skeleton_ctx.queue_state);
free_queue_counter:
	free(skeleton_ctx.queue_counter);
free_program_ctx:
	free(skeleton_ctx.program_ctx);
free_entries:
	free(skeleton_ctx.entries);
	return result;
}

void
flow_skeleton_destroy(void)
{
	int i;

	if (skeleton_ctx.entries != NULL)
		free(skeleton_ctx.entries);

	if (skeleton_ctx.program_ctx != NULL)
		free(skeleton_ctx.program_ctx);

	if (skeleton_ctx.queue_counter != NULL)
		free(skeleton_ctx.queue_counter);

	if (skeleton_ctx.queue_state != NULL) {
		for (i = 0; i < skeleton_ctx.skeleton_cfg->nb_ports; i++) {
			if (skeleton_ctx.queue_state[i] != NULL)
				free(skeleton_ctx.queue_state[i]);
		}
		free(skeleton_ctx.queue_state);
	}
	doca_flow_destroy();
}

/*
 * Add entry based on the given flow_skeleton_entry struct
 *
 * @pipe_queue [in]: queue identifier
 * @flag [in]: DOCA_FLOW_WAIT_FOR_BATCH / DOCA_FLOW_NO_WAIT
 * @entry [in]: application input information for adding the entry
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
add_entry(uint16_t pipe_queue, uint32_t flag, struct flow_skeleton_entry *entry)
{
	doca_error_t result = DOCA_SUCCESS;

	switch (entry->ctx.type) {
	case DOCA_FLOW_PIPE_BASIC:
		result = doca_flow_pipe_add_entry(pipe_queue, entry->ctx.pipe, entry->ctx.match, entry->ctx.actions,
						  entry->ctx.monitor, entry->ctx.fwd, flag, entry->ctx.usr_ctx, entry->ctx.entry);
		break;
	case DOCA_FLOW_PIPE_CONTROL:
		result = doca_flow_pipe_control_add_entry(pipe_queue, entry->ctx.priority, entry->ctx.pipe, entry->ctx.match, entry->ctx.match_mask,
							  entry->ctx.actions, entry->ctx.actions_mask, entry->ctx.action_descs, entry->ctx.monitor, entry->ctx.fwd,
							  entry->ctx.usr_ctx, entry->ctx.entry);
		break;
	case DOCA_FLOW_PIPE_LPM:
		result = doca_flow_pipe_lpm_add_entry(pipe_queue, entry->ctx.pipe, entry->ctx.match, entry->ctx.match_mask,
						      entry->ctx.actions, entry->ctx.monitor, entry->ctx.fwd, flag,
						      entry->ctx.usr_ctx, entry->ctx.entry);
		break;
	case DOCA_FLOW_PIPE_ACL:
		result = doca_flow_pipe_acl_add_entry(pipe_queue, entry->ctx.pipe, entry->ctx.match, entry->ctx.match_mask,
						      entry->ctx.priority, entry->ctx.fwd, flag, entry->ctx.usr_ctx, entry->ctx.entry);
		break;
	case DOCA_FLOW_PIPE_ORDERED_LIST:
		result = doca_flow_pipe_ordered_list_add_entry(pipe_queue, entry->ctx.pipe, entry->ctx.idx, entry->ctx.ordered_list,
							       entry->ctx.fwd, flag, entry->ctx.usr_ctx, entry->ctx.entry);
		break;
	case DOCA_FLOW_PIPE_HASH:
		result = doca_flow_pipe_hash_add_entry(pipe_queue, entry->ctx.pipe, entry->ctx.idx, entry->ctx.actions, entry->ctx.monitor,
							entry->ctx.fwd, flag, entry->ctx.usr_ctx, entry->ctx.entry);
		break;
	default:
		DOCA_LOG_ERR("Unsupported pipe type");
		result = DOCA_ERROR_INVALID_VALUE;
	}
	return result;
}

/*
 * Add a batch of entries and process
 *
 * @params [in]: main loop params
 * @nb_entries [in]: number of entries in the batch
 * @port_id [in]: port ID of the entries
 */
static void
add_batch_entries(struct main_loop_params *params, uint32_t nb_entries, int port_id)
{
	uint32_t flags;
	int i;
	doca_error_t result;

	/* Add all the entries with DOCA_FLOW_WAIT_FOR_BATCH flag, last entry with DOCA_FLOW_NO_WAIT */
	flags = DOCA_FLOW_WAIT_FOR_BATCH;
	for (i = 0; i < nb_entries; i++) {
		if (i == nb_entries - 1)
			flags = DOCA_FLOW_NO_WAIT;
		if (skeleton_ctx.entries[i].ctx.op == DOCA_FLOW_ENTRY_OP_ADD)
			result = add_entry(params->pipe_queue, flags, &skeleton_ctx.entries[i]);
		else if (skeleton_ctx.entries[i].ctx.op == DOCA_FLOW_ENTRY_OP_DEL)
			result = doca_flow_pipe_rm_entry(params->pipe_queue, flags, skeleton_ctx.entries[i].ctx.usr_ctx, *skeleton_ctx.entries[i].ctx.entry);
		else {
			DOCA_LOG_ERR("DOCA Flow op [%d] is not supported", skeleton_ctx.entries[i].ctx.op);
			result = DOCA_ERROR_INVALID_VALUE;
		}
		if (result != DOCA_SUCCESS)
			DOCA_LOG_ERR("Failed to add/remove entry in index [%d]: %s", i, doca_get_error_string(result));
		else
			skeleton_ctx.queue_state[port_id][params->pipe_queue]++;
	}

	result = doca_flow_entries_process(params->ports[port_id], params->pipe_queue, 0, nb_entries);
	if (result != DOCA_SUCCESS)
		DOCA_LOG_ERR("DOCA Flow entries process failed %s", doca_get_error_string(result));

	skeleton_ctx.queue_state[port_id][params->pipe_queue] -= skeleton_ctx.queue_counter[params->pipe_queue];
}

void
flow_skeleton_main_loop(void *main_loop_params)
{
	struct main_loop_params *params = (struct main_loop_params *)main_loop_params;
	uint32_t nb_entries;
	int port_id;
	doca_error_t result;

	skeleton_ctx.program_ctx[params->pipe_queue] = params->program_ctx;

	for (port_id = 0; port_id < params->nb_ports; port_id++) {
		skeleton_ctx.queue_counter[params->pipe_queue] = 0;
		/* Get array of entries to add or remove */
		skeleton_ctx.skeleton_cfg->init_cb(&skeleton_ctx.entries, port_id, params->program_ctx, &nb_entries);
		if (nb_entries == 0)
			continue;
		add_batch_entries(params, nb_entries, port_id);
	}

	while (!force_quit) {
		for (port_id = 0; port_id < params->nb_ports; port_id++) {
			skeleton_ctx.queue_counter[params->pipe_queue] = 0;
			/* Get array of entries to add or remove */
			skeleton_ctx.skeleton_cfg->entries_acquisition_cb(&skeleton_ctx.entries, port_id, params->program_ctx, &nb_entries);
			if (nb_entries == 0)
				continue;
			if (skeleton_ctx.queue_state[port_id][params->pipe_queue] + nb_entries >= skeleton_ctx.queue_depth) {
				DOCA_LOG_WARN("Queue %d on port %d is full", params->pipe_queue, port_id);
				result = doca_flow_entries_process(params->ports[port_id], params->pipe_queue, 0, nb_entries);
				if (result != DOCA_SUCCESS)
					DOCA_LOG_ERR("DOCA Flow entries process failed %s", doca_get_error_string(result));
			}
			add_batch_entries(params, nb_entries, port_id);
		}
	}
}

void
flow_skeleton_notify_exit(void)
{
	force_quit = true;
}
